#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>

#define SCREEN_WIDTH 128
#define SCREEN_HEIGHT 64
#define OLED_RESET -1
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

#define VOLTAGE_PIN A0        // Pin for voltage divider
#define ACS712_PIN A1         // Pin for current sensor (ACS712)
#define RELAY_PIN 7           // Pin to control the relay

const float VREF = 5.0;       // Reference voltage for ADC
const float R1 = 30000.0;     // 30k ohm resistor
const float R2 = 7500.0;      // 7.5k ohm resistor
const float MAX_VOLTAGE = 12.8; // Max voltage for 12.8V LiFePO4
const float UPPER_THRESHOLD = 11.2; // Voltage at which the relay turns on
const float LOWER_THRESHOLD = 11.0; // Voltage at which the relay turns off
const float OVERCURRENT = 12.0; // Overcurrent protection threshold in amps
const int BATTERY_EMPTY = 0;    // Battery empty percentage
const int BATTERY_FULL = 100;   // Battery full percentage

bool relayState = false;  // Keep track of relay state

void setup() {
  pinMode(RELAY_PIN, OUTPUT);
  digitalWrite(RELAY_PIN, LOW);  // Initially turn off the relay

  if(!display.begin(SSD1306_SWITCHCAPVCC, 0x3C)) {
    // If SSD1306 doesn't initialize, stay here
    for(;;);
  }
  display.clearDisplay();
  display.setTextSize(1);
  display.setTextColor(SSD1306_WHITE);
}

float readVoltage() {
  int analogValue = analogRead(VOLTAGE_PIN);
  float voltage = (analogValue * VREF / 1023.0) * ((R1 + R2) / R2);
  return voltage;
}

float calculateBatteryPercentage(float voltage) {
  if(voltage > MAX_VOLTAGE) return BATTERY_FULL;
  if(voltage < LOWER_THRESHOLD) return BATTERY_EMPTY;
  return (voltage - LOWER_THRESHOLD) / (MAX_VOLTAGE - LOWER_THRESHOLD) * 100.0;
}

float readCurrent() {
  int analogValue = analogRead(ACS712_PIN);
  float current = (analogValue - 512) * (VREF / 1023.0) / 0.066; // 66mV/A for ACS712-30A
  return abs(current); // Always return positive current value
}

void drawBatteryIcon(int x, int y, int percentage, bool isCharging) {
  display.drawRect(x, y, 24, 10, SSD1306_WHITE); // Battery outer rectangle
  display.fillRect(x + 24, y + 2, 2, 6, SSD1306_WHITE); // Battery positive terminal

  int fillWidth = map(percentage, 0, 100, 0, 22); // Map percentage to fill width
  display.fillRect(x + 1, y + 1, fillWidth, 8, SSD1306_WHITE); // Fill based on percentage

  if (isCharging) {
    // Draw charging icon inside the battery
    display.drawLine(x + 6, y + 2, x + 11, y + 2, SSD1306_BLACK);
    display.drawLine(x + 11, y + 2, x + 8, y + 8, SSD1306_BLACK);
    display.drawLine(x + 8, y + 8, x + 13, y + 8, SSD1306_BLACK);
  }
}

void updateDisplay(float voltage, float batteryPercentage, float current, bool isOvercurrent, bool isLowBattery) {
  display.clearDisplay();

  // Display battery icon and percentage at the center
  display.setTextSize(2); // Larger text for percentage
  int batteryX = (SCREEN_WIDTH - 24) / 2;
  display.setCursor((SCREEN_WIDTH - 40) / 2, 0);
  display.print((int)batteryPercentage);
  display.println("%");

  drawBatteryIcon(batteryX, 18, (int)batteryPercentage, current < 0);

  // Display voltage and current values with spacing and short labels
  display.setTextSize(1);
  display.setCursor(0, 38); // Start from a lower position
  display.print("V: ");
  display.print(voltage);
  display.println(" V");

  display.setCursor(0, 50); // Next row
  display.print("I: ");
  display.print(current);
  display.println(" A");

  // Handle warnings
  if(isLowBattery) {
    display.clearDisplay();
    display.setTextSize(2);
    display.setCursor(10, 25);
    display.println("LOW BATTERY");
    display.display();
    delay(1500);  // Wait for 1.5 seconds
    display.clearDisplay();
  } else if(isOvercurrent) {
    display.clearDisplay();
    display.setTextSize(2);
    display.setCursor(10, 25);
    display.println("OVERCURRENT");
    display.display();
    delay(1500);  // Wait for 1.5 seconds
    display.clearDisplay();
  }

  display.display();
}

void loop() {
  float voltage = readVoltage();
  float batteryPercentage = calculateBatteryPercentage(voltage);
  float current = readCurrent();

  bool isLowBattery = voltage < LOWER_THRESHOLD;
  bool isOvercurrent = current > OVERCURRENT;

  // Implement hysteresis for relay control with delay for stability
  if (!relayState && voltage > UPPER_THRESHOLD) {
    delay(1500);  // 1.5-second delay to ensure voltage stability
    if (voltage > UPPER_THRESHOLD) { // Check again after delay
      relayState = true;
      digitalWrite(RELAY_PIN, HIGH); // Turn on relay
    }
  } else if (relayState && voltage < LOWER_THRESHOLD) {
    relayState = false;
    digitalWrite(RELAY_PIN, LOW);  // Turn off relay
  }

  if(isOvercurrent) {
    digitalWrite(RELAY_PIN, LOW);  // Turn off relay in case of overcurrent
    relayState = false;
  }

  updateDisplay(voltage, batteryPercentage, current, isOvercurrent, isLowBattery);

  delay(1000); // Delay for a second before next measurement
}
